#ifndef TYPE_CONVERSION_H_BPYMHI7C
#define TYPE_CONVERSION_H_BPYMHI7C

#include <stdint.h>

int8_t to_signed_char(uint8_t x);

int16_t to_signed_short(uint16_t x);

int32_t to_signed_long(uint32_t x);

float to_float(uint32_t x);

#endif /* end of include guard: TYPE_CONVERSION_H_BPYMHI7C */
